/*
 * $Id: UnitTest.cpp,v 1.1 2014-02-10 11:02:24 swaj Exp $
 * Was written by Alexey Semenoff (c) 2012-2014
 * See COPYING for copyiright policy
 */

#include <stdio.h>
#include <sys/time.h>
#include <lzo/lzo1x.h>
#include <algorithm>

/*
 * THIS IS JUST AN EXAMPLE HOW TO USE LIBRARY, NOT COMPILABLE FILE !!! 
 */


//------------------------------------------------------------------------------
static void print_sig(voidp dest, uint len=10);
static void print_string(const char *pfx, uint8_t* s, size_t);

//------------------------------------------------------------------------------
static void print_sig(voidp dest, uint len)
{
    uint8_t* p= (uint8_t*)dest;
    char str[512];

    memset (str, 0x0, sizeof(str));
    snprintf (str, sizeof(str)-1, "Signature: ");

    for (size_t i=0; i< len; i++) {
	snprintf (str+strlen(str), sizeof(str)-1-strlen(str), " 0x%x", p[i] );
    }

    snprintf (str+strlen(str), sizeof(str)-1-strlen(str), " [ ");

    for (size_t i=0; i< len; i++) {
        if (p[i]>31) {
	    snprintf (str+strlen(str), sizeof(str)-1-strlen(str), " %c", p[i]);
        }
        else {
	    snprintf (str+strlen(str), sizeof(str)-1-strlen(str), " ?");
        }
    }

    snprintf (str+strlen(str), sizeof(str)-1-strlen(str), " ]");
    Info("%s", str);
}

//------------------------------------------------------------------------------
static void print_string(const char *pfx, uint8_t* s, size_t len)
{
  const size_t maxExpectedLen = 2048;
  const size_t maxPrintable = 32;

  if (len > maxExpectedLen) {
    throw error("!!! very strange big data(>2K), panic!!!");
  }
  char* str2 = (char*)malloc(len + 1);
  if (! str2) {
    throw error("no memory");
  }
  memcpy((void *)str2, (const void *)s, len+1);
  if (len > maxPrintable) {
    str2[maxPrintable] = 0x0;
  }

  Info ("%s '%s%s [%zu bytes]", pfx, str2, (len>maxPrintable) ? "... '" : "'", len );
  free(str2);
}

//------------------------------------------------------------------------------
typedef CLzoStream::lzo_stream lzo_stream;

static lzo_stream zs;
static const uint buf_sz=(1024 * 80);
static const uint data_len = (1024 * 70 + 101);
static uint8_t orig_buf[buf_sz];
static uint8_t dest_buf[buf_sz*2];
static uint orig_processed = 0;
static uint dest_processed = 0;

//------------------------------------------------------------------------------
static int lzoDeflateRev (lzo_stream *s, int mode)
{
    uint32_t av_out = s->avail_out;
    int rc = CLzoStream::lzoDeflate(s, mode);
    s->avail_out = av_out - s->avail_out;
    return rc;
}

//------------------------------------------------------------------------------
static int lzoInflateRev (lzo_stream *s, int mode)
{
    uint32_t av_out = s->avail_out;
    int rc = CLzoStream::lzoInflate(s, mode);
    s->avail_out = av_out - s->avail_out;
    return rc;
}

//------------------------------------------------------------------------------
static
void lzo_compress_whole_buff ()
{
    Info ("--- lzo_compress_whole_buff {");

    memset (&zs, 0x0, sizeof(lzo_stream));
    ASSERT_EQUAL(int, CLzoStream::lzoDeflateInit(&zs), CLzoStream::EOk); 

    zs.next_in = orig_buf;
    zs.avail_in = data_len;

    zs.next_out = dest_buf;
    zs.avail_out = sizeof(dest_buf);

    Info("! CLzoStream::lzoDeflate {");
    ASSERT_EQUAL(int, lzoDeflateRev(&zs, CLzoStream::ELzoSyncFlush), CLzoStream::EStreamEnd ); 
    Info("! } CLzoStream::lzoDeflate");
    Info("avail_in= %d, avail_out= %d", zs.avail_in, zs.avail_out);  

    ASSERT_EQUAL(int, zs.avail_in, 0);
    ASSERT_EQUAL(int, zs.total_in, data_len);
    ASSERT_EQUAL(int, zs.total_out, zs.avail_out);

    dest_processed = zs.avail_out;

    ASSERT_EQUAL(int, CLzoStream::lzoDeflateEnd(&zs), CLzoStream::EOk); 

    Info ("} lzo_compress_whole_buff ---");
}

//-------------------------------------------------------------------------------
static
void lzo_analyze_chunks ()
{
    Info ("--- lzo_analyze_chunks {");

    uint8_t* destp = dest_buf;

    int chunkno=0;
    uint32_t tot_comp=0;
    uint32_t tot_uncomp=0;
    uint8_t sig;
    uint32_t comp_sz, uncomp_sz, crc;

    while (1)
    {
	sig= *destp;
        comp_sz = *((uint32_t*)(destp+1));
        uncomp_sz = *((uint32_t*)(destp+5));
        crc = *((uint32_t*)(destp+9));

	Info("chunk #%d, sig= 0x%x, c_len= %d, u_len= %d, crc= 0x%x",
	     chunkno++, sig, comp_sz, uncomp_sz, crc );
	print_sig (destp,CLzoStream::const_HEADER_LEN);


	ASSERT_EQUAL(int, sig, CLzoStream::ESigLzo1xAdler); 
	
        tot_comp += comp_sz;
        tot_uncomp += uncomp_sz;

	destp += (comp_sz + CLzoStream::const_HEADER_LEN);

	ASSERT( destp < (dest_buf + dest_processed + 1 ) );

	if ( destp >= (dest_buf + dest_processed) ) {
	    break;
	}
    }

    Info ("chunks= %d, tot_comp=%d [%d], tot_uncomp= %d",
	  chunkno, tot_comp, (tot_comp+CLzoStream::const_HEADER_LEN * chunkno), tot_uncomp );

    ASSERT_EQUAL(int, dest_processed , tot_comp+CLzoStream::const_HEADER_LEN * chunkno); 
    Info ("} lzo_analyze_chunks ---");
}


//------------------------------------------------------------------------------
static
void lzo_compress_by_bytes_no_flush (uint len)
{
   Info ("--- compress_by_bytes_no_flush {");

   memset (&zs, 0x0, sizeof(lzo_stream));
   ASSERT_EQUAL(int, CLzoStream::lzoDeflateInit(&zs), CLzoStream::EOk); 

   int full_blocks = data_len / len;
   int blocks_tail = data_len % len;

   Info ("pack %d by %d bytes, we have %d fill block + %d tail",
	 data_len, len, full_blocks, blocks_tail);

   orig_processed = 0;
   dest_processed = 0;

   // pack N blocks
   for (int bl=0; bl< full_blocks; bl++)
   {
       zs.next_in = orig_buf + orig_processed;
       zs.avail_in = len;

       zs.next_out = dest_buf + dest_processed;
       zs.avail_out = buf_sz - dest_processed;

       ASSERT_EQUAL(int, lzoDeflateRev (&zs, CLzoStream::ELzoNoFlush), CLzoStream::EOk);
       ASSERT_EQUAL(int, zs.avail_in, 0);

       if (zs.avail_out) {
	   Info("Output data ready %d bytes", zs.avail_out);
       }

       dest_processed += zs.avail_out;
       orig_processed += len;
   }

   // Flush buffer with tail
   zs.next_in = orig_buf + orig_processed;
   zs.avail_in = blocks_tail;

   zs.next_out = dest_buf + dest_processed;
   zs.avail_out = buf_sz - dest_processed;

   ASSERT_EQUAL(int, lzoDeflateRev (&zs, CLzoStream::ELzoFullFlush), CLzoStream::EStreamEnd);

   dest_processed += zs.avail_out;

   ASSERT_EQUAL(int, zs.avail_in, 0);
   ASSERT_EQUAL(int, zs.total_out, dest_processed);
   ASSERT_EQUAL(int, zs.total_in, data_len);
   ASSERT_EQUAL(int, CLzoStream::lzoDeflateEnd(&zs), CLzoStream::EOk); 

   dest_processed += zs.avail_out;

   Info("Compress, %d -> %d", data_len, dest_processed);
   Info ("} lzo_decompress_by_bytes ---");
}

//------------------------------------------------------------------------------
static
void lzo_decompress_by_bytes (uint len)
{
   Info ("--- lzo_decompress_by_bytes {");

   memset (&zs, 0x0, sizeof(lzo_stream));
   ASSERT_EQUAL(int, CLzoStream::lzoDeflateInit(&zs), CLzoStream::EOk); 

   int full_blocks = dest_processed / len;
   int blocks_tail = dest_processed % len;

   Info ("unpack %d by %d bytes, we have %d fill block + %d tail",
	 dest_processed, len, full_blocks, blocks_tail);

    uint8_t my_dest[buf_sz*2];
    memset (my_dest, 0x0, sizeof(my_dest));

   int my_orig_processed = 0;
   int my_dest_processed = 0;

   int zip_rv;

   // unpack N blocks
   for (int bl=0; bl< full_blocks; bl++)
   {
       zs.next_in = dest_buf + my_orig_processed;
       zs.avail_in = len;

       zs.next_out = my_dest + my_dest_processed;
       zs.avail_out = sizeof(my_dest) - my_dest_processed;

       
       zip_rv = lzoInflateRev (&zs, CLzoStream::ELzoNoFlush);
       if (zip_rv != CLzoStream::EOk && zip_rv != CLzoStream::EStreamEnd ) {
	   Info ("zip_rv= %d, must be %d or %d",
		 zip_rv,  CLzoStream::EOk, CLzoStream::EStreamEnd );
	   ASSERT(0);
       }

       ASSERT_EQUAL(int, zs.avail_in, 0);

       if (zs.avail_out) {
	   Info("Output data ready %d bytes", zs.avail_out);
	   my_dest_processed += zs.avail_out;

	   Info("processed so far... %d", my_dest_processed);

	   if ( memcmp((const void *)my_dest, (const void *)orig_buf, my_dest_processed) !=0  ) {
	       
	       ASSERT(0);
	   }

	   Info("Compared %d bytes with orig_buf OK", my_dest_processed);

       }

       my_orig_processed += len;
   }

   // Unpack tail if any
   if (blocks_tail) {

       zs.next_in = dest_buf + my_orig_processed;
       zs.avail_in = blocks_tail;

       zs.next_out = my_dest + my_dest_processed;
       zs.avail_out = sizeof(my_dest) - my_dest_processed;

       ASSERT_EQUAL(int, lzoInflateRev (&zs, CLzoStream::ELzoFullFlush), CLzoStream::EStreamEnd);

       my_dest_processed += zs.avail_out;

       Info("my_dest_processed= %d", my_dest_processed);
   }

   ASSERT_EQUAL(int, zs.avail_in, 0);
   ASSERT_EQUAL(int, zs.total_out, my_dest_processed);
   ASSERT_EQUAL(int, zs.total_in, dest_processed);
   ASSERT_EQUAL(int, CLzoStream::lzoDeflateEnd(&zs), CLzoStream::EOk); 

   Info ("Compare uncompressed data with original len=%d", data_len);
   ASSERT_EQUAL(int, memcmp((const void *)my_dest, (const void *)orig_buf, data_len ), 0);
   Info ("Data is the same");

   Info ("} lzo_decompress_by_bytes ---");
}

//------------------------------------------------------------------------------
static
void lzo_decompress_whole()
{
    Info ("--- lzo_decompress_whole {");

    memset (&zs, 0x0, sizeof(lzo_stream));
    ASSERT_EQUAL(int, CLzoStream::lzoInflateInit(&zs), CLzoStream::EOk); 

    zs.next_in = dest_buf;
    zs.avail_in = dest_processed;

    uint8_t my_dest[buf_sz*2];
    memset (my_dest, 0x0, sizeof(my_dest));

    zs.next_out = my_dest;
    zs.avail_out = sizeof(my_dest);

    Info("! CLzoStream::lzoInflate {");
    ASSERT_EQUAL(int, lzoInflateRev(&zs, CLzoStream::ELzoSyncFlush), CLzoStream::EStreamEnd); 
    Info("! } CLzoStream::lzoInflate");
    Info("avail_in= %d, avail_out= %d", zs.avail_in, zs.avail_out);  

    ASSERT_EQUAL(int, zs.avail_in, 0);
    ASSERT_EQUAL(int, zs.total_out, data_len);
    ASSERT_EQUAL(int, zs.avail_out, data_len);
    ASSERT_EQUAL(int, CLzoStream::lzoDeflateEnd(&zs), CLzoStream::EOk); 
    Info ("Compare uncompressed data with original");
    ASSERT_EQUAL(int, memcmp((const void *)my_dest, (const void *)orig_buf  , data_len), 0);
    Info ("Data is the same");
    Info ("} lzo_decompress_whole ---");
}

//------------------------------------------------------------------------------
void DataCompressorTest::verifyStreamCompression()
{
    Info ("Stream compression/decompression");
    printf ("Stream compression/decompression\n");

    bool dump_chunks = false;

    ASSERT (data_len < buf_sz);

    uint8_t mask = 0xff;
    bool do_it=true;

    do {
	Info ("0x%x", mask);

	if (! mask) {
	    do_it = false;
	}

	struct timeval tv;
	gettimeofday(&tv, NULL);
	uint salt = getpid() ^ tv.tv_usec ^ (uint)tv.tv_sec;
	Info ("srand(%d)", salt);
	printf ("srand(%d) mask= 0x%x\n", salt, mask);
	srand(salt);
    
	Info ("Populating rand() buffer, %d of %d bytes", data_len, buf_sz);  


	for (uint i=0; i< data_len; i++)
	{
	    orig_buf[i] = ( (uint8_t)rand() & mask);
	}

	Info ("Buffer filled");
	print_sig (orig_buf);

	// compress/decompress at once
	lzo_compress_whole_buff();

	uint my_compressed_len = dest_processed;

	Info ("Compress ratio is %2.f%%", (double) my_compressed_len / (double)data_len * 100.0);
	if (dump_chunks) {
	    lzo_analyze_chunks ();
	}

	lzo_decompress_whole();

	// compress by pieces
	for (uint unit_len=1; unit_len <=17000; unit_len +=5) {
	    Info ("Compress by %d", unit_len);
	    dest_processed = 0;
	    lzo_compress_by_bytes_no_flush (unit_len);
	    ASSERT_EQUAL(int, dest_processed, my_compressed_len); 
	    if (dump_chunks) {
		lzo_analyze_chunks ();
	    }
	    lzo_decompress_whole();
	}

	// compress by 1, decompress at once
	lzo_compress_whole_buff();
	ASSERT_EQUAL(int, dest_processed, my_compressed_len);
	
	if (dump_chunks) {
	    lzo_analyze_chunks ();
	}

	// decompress by pieces
	for (uint unit_len=1; unit_len <= my_compressed_len; unit_len+=10) {
	    Info ("Decmpress by %d", unit_len);
	    lzo_decompress_by_bytes (unit_len);
	    
	}

	mask >>= 1;
	mask >>= 1;

    } while ( do_it );
}

//------------------------------------------------------------------------------
static
void lzo_compress_whole_buff_by_pieces (uint piece)
{
    Info ("--- lzo_compress_whole_buff by %d {", piece );

    memset (&zs, 0x0, sizeof(lzo_stream));
    ASSERT_EQUAL(int, CLzoStream::lzoDeflateInit(&zs), CLzoStream::EOk); 

    zs.next_in = orig_buf;
    zs.avail_in = data_len;

    zs.next_out = dest_buf;
    zs.avail_out = piece;

    dest_processed = 0;

    while (1)
    {
	Info("! CLzoStream::lzoDeflate {");
	ASSERT_EQUAL(int, lzoDeflateRev(&zs, CLzoStream::ELzoNoFlush ), CLzoStream::EOk); 
	Info("! } CLzoStream::lzoDeflate");
	Info("avail_in= %d, avail_out= %d", zs.avail_in, zs.avail_out);  

	if (! zs.avail_out &&  ! zs.avail_in ) {
	    break;
	}

	dest_processed +=  zs.avail_out;
	zs.next_out = dest_buf + dest_processed ;
	//Info ("zs.next_out = %p, zs.avail_out= %d, zs.next_out-dest_buf =%d",
	//      zs.next_out,  zs.avail_out, zs.next_out-dest_buf);


	zs.avail_out = piece;
    }

    Info(" --- FLUSH ---\n");

    while (1)
    {
	zs.avail_out = piece;

	Info("! CLzoStream::lzoDeflate FLUSH {");
	
	int zip_rv = lzoDeflateRev(&zs, CLzoStream::ELzoSyncFlush );
	if (zip_rv != CLzoStream::EStreamEnd && zip_rv != CLzoStream::EOk)
	{
	    ASSERT(0);
	}

	Info("! } CLzoStream::lzoDeflate FLUSH");

	if ( ! zs.avail_out && !zs.avail_in ) {
	    break;
	}

	dest_processed +=  zs.avail_out;
	zs.avail_out = piece;

    }

    Info ("} lzo_compress_whole_buff by %d ---", piece);
}

//------------------------------------------------------------------------------
static
void lzo_decompress_whole_by_pieces(uint piece)
{
    Info ("--- lzo_decompress_whole by %d {", piece);

    memset (&zs, 0x0, sizeof(lzo_stream));
    ASSERT_EQUAL(int, CLzoStream::lzoInflateInit(&zs), CLzoStream::EOk); 

    zs.next_in = dest_buf;
    zs.avail_in = dest_processed;

    uint8_t my_dest[buf_sz*2];
    memset (my_dest, 0x0, sizeof(my_dest));

    zs.next_out = my_dest;
    zs.avail_out = piece;

    int dpr=0;
    int zip_rv;

    while (1)
    {
	Info("! CLzoStream::lzoInflate {");

	zip_rv = lzoInflateRev(&zs, CLzoStream::ELzoSyncFlush);

	if (zip_rv != CLzoStream::EStreamEnd && zip_rv != CLzoStream::EOk) {
	    Info ("zip_rv = %d", zip_rv);
	    ASSERT(0);
	}

	Info("! } CLzoStream::lzoInflate");
	Info("avail_in= %d, avail_out= %d", zs.avail_in, zs.avail_out);  

	if (! zs.avail_out &&  ! zs.avail_in ) {
	    break;
	}

	dpr +=  zs.avail_out;
	zs.next_out = my_dest + dpr;
	zs.avail_out = piece;
    }

    Info ("Uncompressed %d bytes", dpr);

    ASSERT_EQUAL(int, zs.avail_in, 0);
    ASSERT_EQUAL(int, zs.total_out, data_len);
    ASSERT_EQUAL(int, dpr, data_len);

    ASSERT_EQUAL(int, CLzoStream::lzoDeflateEnd(&zs), CLzoStream::EOk); 

    Info ("Compare uncompressed data with original");
    ASSERT_EQUAL(int, memcmp((const void *)my_dest, (const void *)orig_buf  , data_len), 0);
    Info ("Data is the same");
    Info ("} lzo_decompress_whole by %d ---", piece);
}


//-------------------------------------------------------------------------------
void DataCompressorTest::verifyStreamCompressionSmallBuf()
{
    Info ("Stream compression/decompression ***SMALL BUF***");
    printf ("Stream compression/decompression ***SMALL BUF***\n");

    bool dump_chunks = false;

    ASSERT (data_len < buf_sz);

    uint8_t mask = 0x3;

    do {
	struct timeval tv;
	gettimeofday(&tv, NULL);
	uint salt = getpid() ^ tv.tv_usec ^ (uint)tv.tv_sec;
	Info ("srand(%d)", salt);
	printf ("srand(%d) mask= 0x%x\n", salt, mask);
	srand(salt);

	//srand(0);
    
	Info ("Populating rand() buffer, %d of %d bytes", data_len, buf_sz);  

	for (uint i=0; i< data_len; i++)
	{
	    orig_buf[i] = ( (uint8_t)rand() & mask);
	}

	Info ("Buffer filled");
	print_sig (orig_buf);

	lzo_compress_whole_buff();

	int my_dest_processed = dest_processed;

	for (uint pcs=1; pcs< 10000; pcs += 999) {
	    printf ("lzo_compress by %d\n", pcs);
	    lzo_compress_whole_buff_by_pieces(pcs);
	    ASSERT_EQUAL(int, dest_processed, my_dest_processed); 
	}

	Info ("dest_processed= %d\n", dest_processed );

	if (dump_chunks) {
	    lzo_analyze_chunks ();
	}

	lzo_decompress_whole();


	for (uint pcs=1; pcs< 10000; pcs += 999) {
	    printf ("lzo_decompress by %d\n", pcs);
	    lzo_decompress_whole_by_pieces(pcs);
	}


    } while ( 0 );
}
//-------------------------------------------------------------------------------


//__END__
