/*
 * $Id: liblzostream.h,v 1.2 2014-02-10 10:28:29 swaj Exp $
 * Was written by Alexey Semenoff (c) 2012-2014
 * See COPYING for copyiright policy
 */

#ifndef __LIBLZOSTREAM_H__
#define __LIBLZOSTREAM_H__

#include <stdio.h>
#include <inttypes.h>
#include "liblzostream_utils.h"

//------------------------------------------------------------------------------
class CLzoStream : public SDataCompressorUtils
{
// ------------------- Interface --------
public:
    typedef SDataCompressorUtils::compressor_stream lzo_stream;
    static int lzoInflateInit   (lzo_stream *s);
    static int lzoInflate       (lzo_stream *s, int mode);
    static int lzoInflateReset  (lzo_stream *s);
    static int lzoInflateEnd    (lzo_stream *s);
    static int lzoDeflateInit   (lzo_stream *s);
    static int lzoDeflate       (lzo_stream *s, int mode);
    static int lzoDeflateReset  (lzo_stream *s);
    static int lzoDeflateEnd    (lzo_stream *s);
    static int lzoGetPageSize() { return stream_PAGE_SIZE; };
    static int lzoGetPageSize(lzo_stream *s);
    static int lzoSetPageSize(unsigned int psize);
    static int lzoSetPageSize(lzo_stream *s, unsigned int psize);

    static int lzoInflateCached (lzo_stream *s);
    static const uint8_t* lzoInflateCachedPtr (lzo_stream *s);
    static int lzoFlushIntCache (lzo_stream *s);

public:
    static int lzoInflateInit  (lzo_stream& s) { return lzoInflateInit(&s); };
    static int lzoInflate      (lzo_stream& s, int mode) { return lzoInflate (&s, mode); };
    static int lzoInflateReset (lzo_stream& s) { return lzoInflateReset (&s); };
    static int lzoInflateEnd   (lzo_stream& s) { return lzoInflateEnd(&s); };
    static int lzoDeflateInit  (lzo_stream& s) { return lzoDeflateInit(&s); };
    static int lzoDeflate      (lzo_stream& s, int mode) { return lzoDeflate(&s, mode); };
    static int lzoDeflateReset (lzo_stream& s) { return  lzoDeflateReset(&s); };
    static int lzoDeflateEnd   (lzo_stream& s) { return lzoDeflateEnd(&s); };
// --------------------------------------

public:
    static const unsigned int const_HEADER_LEN;
    static unsigned int stream_PAGE_SIZE;
    static const unsigned int const_PAGE_SIZE_MIN;
    static const unsigned int const_PAGE_SIZE_MAX;
private:
    static const int const_CRC_INITIAL;
    static const unsigned int const_SIGLEN;
    static const unsigned int const_INT_BUF_SIZE;
    static const uint8_t const_ESigLzo1xAdler;

    struct internal_state : stream_internal_state {
        voidp wrkmem;
        internal_state() {
            wrkmem = 0;
        }
    };

private:
    static int pack_chunk_from_stream (lzo_stream *s, size_t bytes_to_pack);
    static int pack_chunk_from_cache (lzo_stream *s, size_t bytes_to_pack, size_t bytes_to_report);
    static int pack_chunks_noflush (lzo_stream *s);
    static int pack_chunks_flush (lzo_stream *s);
};

//------------------------------------------------------------------------------

#endif //__LIBLZOSTREAM_H__
