/*
 * $Id: liblzostream_utils.h,v 1.1 2014-02-10 10:28:29 swaj Exp $
 * Was written by Alexey Semenoff (c) 2012-2014
 * See COPYING for copyiright policy
 */

//------------------------------------------------------------------------------
#ifndef __LIBLZOSTREAM_UTILS_H__
#define __LIBLZOSTREAM_UTILS_H__
//------------------------------------------------------------------------------

#include <lzo/lzo1x.h>
#include <stdint.h>

//------------------------------------------------------------------------------
#ifndef __VOIDP_DEFINED__
#define __VOIDP_DEFINED__
typedef void* voidp;
#endif // __VOIDP_DEFINED__

//------------------------------------------------------------------------------
class SDataCompressorUtils
{
public:
    enum PackAlgo {
        ENone     = 0,
        ELzo      = 1,
        ELzoRaw   = 2,
        EInvalid  = 255
    };

    enum Signature {
        ESigNone        = 0,
        ESigNoCompAdler = 0x9,
        ESigLzo1xAdler  = 0x10,
        ESigLzo1xNoCrc  = 0x11,
        ESigMax         = 0xFF
    };

    enum Error {
        // DataCompressor
        ENoerror           = 0,
        EOk                = 0,
        EError             = -1,
        ENotSupported      = 254,
        ENotInited         = 255,
        EOutputBufFull     = 71,

        // Stream
        EStreamEnd         = 1,
        ENotCompressible   = 3,
        EStreamError       = -12,
        EDataError         = -13,
        EMemError          = -14,
        EBufError          = -15,
        EVersionError      = -16,

        // Lzo
        EInputOverrun      = -24,
        EOutputOverrun     = -25,
        ELookBehOverrun    = -26,
        EEofNotFound       = -27,
        EInputNotConsumed  = -28,
   };

    // for streams
    enum EMode {
        ELzoNoFlush      = 0, // lzo + zlib
        ELzoPartialFlush = 2, // zlib, deprecated
        ELzoSyncFlush    = 2, // lzo + zlib
        ELzoFullFlush    = 3, // lzo + zlib
        ELzoFinish       = 4, // lzo + zlib
        ELzoBlock        = 5,  // zlib
        ENoCompNoFlush   = 0,
        ENoCompSyncFlush = 2,
        ENoCompFinish    = 4
    };

protected:

    struct stream_internal_state {
        uint8_t  *in_buffer;
        uint32_t av_in;
        uint8_t  *out_buffer;
        uint32_t av_out;
        uint32_t av_out_cached;
        uint32_t page_size;
        void reset() {
            av_in = av_out = av_out_cached = 0;
        }
        void set_page_size (uint32_t sz) { page_size = sz; }
        uint32_t get_page_size() const { return page_size; }
        stream_internal_state() {
            in_buffer = out_buffer = 0;
            page_size = 0;
            reset();
        };
    };

    struct compressor_stream { // same shit as zlib stream
        uint8_t  *next_in;  /* next input byte */
        uint32_t avail_in;  /* number of bytes available at next_in */
        size_t   total_in;  /* total nb of input bytes read so far */
        uint8_t  *next_out; /* next output byte should be put there */
        uint32_t avail_out; /* remaining free space at next_out */
        size_t   total_out; /* total nb of bytes output so far */
        voidp    int_st;    /* not visible by applications */
        inline void update_avail_out (uint32_t av_out) { avail_out = av_out - avail_out; }
        void reset() {
            next_in=0;
            avail_in=0;
            total_in=0;
            next_out=0;
            avail_out=0;
            total_out=0;
        }
        void reset_stat() {
            total_in=0;
            total_out=0;
        }
        compressor_stream () {
            int_st = 0;
            reset();
        }
    };

    static void prepare_stream (compressor_stream *s, stream_internal_state *st, uint32_t& av_out){
        st->av_out = s->avail_out;
        av_out = s->avail_out;
        s->avail_out = 0;
    }

private:
    // low level lib error mapping
    typedef struct {
        int zip_err;
        int dc_err;
    } ErrMap;

    static const ErrMap SLzoErrors[];  // LZO one

public:
    static int getError(PackAlgo algo, int errno);
};
//------------------------------------------------------------------------------

#endif // __LIBLZOSTREAM_UTILS_H__
